"""Tasmota light."""
from __future__ import annotations

import colorsys
from dataclasses import dataclass
import logging
from typing import Any, cast

from .const import (
    COMMAND_CHANNEL,
    COMMAND_COLOR,
    COMMAND_CT,
    COMMAND_DIMMER,
    COMMAND_FADE,
    COMMAND_POWER,
    COMMAND_SCHEME,
    COMMAND_SPEED,
    COMMAND_WHITE,
    CONF_LIGHT_SUBTYPE,
    CONF_LINK_RGB_CT,
    CONF_MAC,
    CONF_OPTIONS,
    CONF_RELAY,
    CONF_TUYA,
    LST_COLDWARM,
    LST_NONE,
    LST_RGB,
    LST_RGBCW,
    LST_RGBW,
    LST_SINGLE,
    OPTION_FADE_FIXED_DURATION,
    OPTION_NOT_POWER_LINKED,
    OPTION_PWM_MULTI_CHANNELS,
    OPTION_REDUCED_CT_RANGE,
    RL_LIGHT,
)
from .entity import (
    TasmotaAvailability,
    TasmotaAvailabilityConfig,
    TasmotaEntity,
    TasmotaEntityConfig,
)
from .mqtt import ReceiveMessage, send_commands
from .utils import (
    config_get_friendlyname,
    config_get_state_offline,
    config_get_state_online,
    config_get_state_power_off,
    config_get_state_power_on,
    get_state_power,
    get_topic_command,
    get_topic_command_state,
    get_topic_stat_result,
    get_topic_tele_state,
    get_topic_tele_will,
    get_value_by_path,
)

LIGHT_TYPE_NONE = 0
LIGHT_TYPE_DIMMER = 1
LIGHT_TYPE_COLDWARM = 2
LIGHT_TYPE_RGB = 3
LIGHT_TYPE_RGBW = 4
LIGHT_TYPE_RGBCW = 5

LIGHT_TYPE_MAP = {
    LIGHT_TYPE_NONE: LST_NONE,
    LIGHT_TYPE_DIMMER: LST_SINGLE,
    LIGHT_TYPE_COLDWARM: LST_COLDWARM,
    LIGHT_TYPE_RGB: LST_RGB,
    LIGHT_TYPE_RGBW: LST_RGBW,
    LIGHT_TYPE_RGBCW: LST_RGBCW,
}

DEFAULT_MIN_MIREDS = 153
DEFAULT_MAX_MIREDS = 500
REDUCED_MIN_MIREDS = 200
REDUCED_MAX_MIREDS = 380

_LOGGER = logging.getLogger(__name__)


@dataclass(frozen=True, kw_only=True)
class TasmotaLightConfig(TasmotaAvailabilityConfig, TasmotaEntityConfig):
    """Tasmota light configuation."""

    idx: int

    dimmer_cmd: str
    dimmer_state: str
    color_suffix: str
    command_topic: str
    control_by_channel: bool
    fade_fixed_duration: bool
    light_type: int
    max_mireds: int
    min_mireds: int
    not_power_linked: bool
    poll_topic: str
    result_topic: str
    state_power_off: str
    state_power_on: str
    state_topic: str
    tuya: bool

    @classmethod
    def from_discovery_message(
        cls, config: dict, idx: int, platform: str
    ) -> TasmotaLightConfig:
        """Instantiate from discovery message."""
        color_suffix = ""
        dimmer_cmd = COMMAND_DIMMER
        dimmer_state = COMMAND_DIMMER
        control_by_channel = False  # Use Channel<n> command to control the light
        if config[CONF_TUYA]:
            dimmer_idx = 3  # Brightness controlled by DIMMER3
            dimmer_cmd = f"{COMMAND_DIMMER}{dimmer_idx}"
        tasmota_light_sub_type = config[CONF_LIGHT_SUBTYPE]
        light_type = LIGHT_TYPE_MAP[tasmota_light_sub_type]

        if config[CONF_OPTIONS][OPTION_PWM_MULTI_CHANNELS]:
            # Multi-channel PWM instead of a single light, each light controlled by CHANNEL<n>
            dimmer_state = f"{COMMAND_CHANNEL}{idx+1}"
            control_by_channel = True
            light_type = LIGHT_TYPE_DIMMER
        elif not config[CONF_LINK_RGB_CT] and tasmota_light_sub_type >= LST_RGBW:
            # Split light in RGB (idx==0) + White/CT (idx==1)
            first_light = config[CONF_RELAY].index(RL_LIGHT)
            if idx - first_light == 0:
                dimmer_idx = 1  # Brightness controlled by DIMMER1
                light_type = LIGHT_TYPE_RGB
                color_suffix = "="
            if idx - first_light == 1:
                dimmer_idx = 2  # Brightness controlled by DIMMER2
                if tasmota_light_sub_type == LST_RGBW:
                    light_type = LIGHT_TYPE_DIMMER
                else:
                    light_type = LIGHT_TYPE_COLDWARM
            dimmer_cmd = f"{COMMAND_DIMMER}{dimmer_idx}"
            dimmer_state = f"{COMMAND_DIMMER}{dimmer_idx}"

        min_mireds = DEFAULT_MIN_MIREDS
        max_mireds = DEFAULT_MAX_MIREDS
        if config[CONF_OPTIONS][OPTION_REDUCED_CT_RANGE]:
            min_mireds = REDUCED_MIN_MIREDS
            max_mireds = REDUCED_MAX_MIREDS

        return cls(
            endpoint="light",
            idx=idx,
            friendly_name=config_get_friendlyname(config, platform, idx),
            mac=config[CONF_MAC],
            platform=platform,
            poll_payload="",
            poll_topic=get_topic_command_state(config),
            availability_topic=get_topic_tele_will(config),
            availability_offline=config_get_state_offline(config),
            availability_online=config_get_state_online(config),
            dimmer_cmd=dimmer_cmd,
            dimmer_state=dimmer_state,
            color_suffix=color_suffix,
            command_topic=get_topic_command(config),
            control_by_channel=control_by_channel,
            fade_fixed_duration=config[CONF_OPTIONS][OPTION_FADE_FIXED_DURATION],
            light_type=light_type,
            max_mireds=max_mireds,
            min_mireds=min_mireds,
            not_power_linked=config[CONF_OPTIONS][OPTION_NOT_POWER_LINKED],
            result_topic=get_topic_stat_result(config),
            state_power_off=config_get_state_power_off(config),
            state_power_on=config_get_state_power_on(config),
            state_topic=get_topic_tele_state(config),
            tuya=config[CONF_TUYA],
        )


class TasmotaLight(TasmotaAvailability, TasmotaEntity):
    """Representation of a Tasmota light."""

    _cfg: TasmotaLightConfig

    def __init__(self, **kwds: Any):
        """Initialize."""
        self._brightness = None
        self._color: list[float] | None = None
        self._color_temp: int | None = None
        self._state: bool | None = None
        self._sub_state: dict | None = None
        super().__init__(**kwds)

    async def subscribe_topics(self) -> None:
        """Subscribe to topics."""

        def state_message_received(msg: ReceiveMessage) -> None:
            """Handle new MQTT state messages."""
            if not self._on_state_callback:
                return

            attributes = {}
            idx = self._cfg.idx

            if self._cfg.endpoint == "light":
                if self._cfg.light_type != LIGHT_TYPE_NONE:
                    brightness = get_value_by_path(
                        msg.payload, [self._cfg.dimmer_state]
                    )
                    if brightness is not None:
                        self._brightness = brightness
                        attributes["brightness"] = brightness

                    if (
                        color := get_value_by_path(msg.payload, [COMMAND_COLOR])
                    ) is not None:
                        if color.find(",") != -1:
                            color = color.split(",", 3)
                        else:
                            color = [
                                int(color[i : i + 2], 16)
                                for i in range(0, len(color), 2)
                            ]
                        if len(color) >= 3:
                            color = [float(color[0]), float(color[1]), float(color[2])]
                            self._color = color
                            attributes["color"] = color

                    if (
                        color_hsb := get_value_by_path(msg.payload, ["HSBColor"])
                    ) is not None:
                        color_hsb = color_hsb.split(",", 3)
                        if len(color_hsb) == 3:
                            color_hs = [float(color_hsb[0]), float(color_hsb[1])]
                            attributes["color_hs"] = color_hs

                    if (
                        color_temp := get_value_by_path(msg.payload, [COMMAND_CT])
                    ) is not None:
                        self._color_temp = color_temp
                        attributes["color_temp"] = color_temp

                    scheme = get_value_by_path(msg.payload, [COMMAND_SCHEME])
                    if scheme is not None and self.effect_list:
                        try:
                            attributes["effect"] = self.effect_list[scheme]
                        except IndexError:
                            attributes["effect"] = f"Scheme {scheme}"
                            _LOGGER.debug("Unknown scheme %s", scheme)

                    if (
                        white_value := get_value_by_path(msg.payload, [COMMAND_WHITE])
                    ) is not None:
                        attributes["white_value"] = white_value

            state = get_state_power(cast(str, msg.payload), idx)

            if state == self._cfg.state_power_on:
                self._state = True
                self._on_state_callback(True, attributes=attributes)
            elif state == self._cfg.state_power_off:
                self._state = False
                self._on_state_callback(False, attributes=attributes)

        availability_topics = self.get_availability_topics()
        topics = {
            "result_topic": {
                "event_loop_safe": True,
                "topic": self._cfg.result_topic,
                "msg_callback": state_message_received,
            },
            "state_topic": {
                "event_loop_safe": True,
                "topic": self._cfg.state_topic,
                "msg_callback": state_message_received,
            },
        }
        topics = {**topics, **availability_topics}

        self._sub_state = await self._mqtt_client.subscribe(
            self._sub_state,
            topics,
        )

    async def unsubscribe_topics(self) -> None:
        """Unsubscribe to all MQTT topics."""
        self._sub_state = await self._mqtt_client.unsubscribe(self._sub_state)

    @property
    def effect_list(self) -> list[str] | None:
        """Return effect list."""
        if self._cfg.endpoint == "light":
            return ["Solid", "Wake up", "Cycle up", "Cycle down", "Random"]
        return None

    @property
    def light_type(self) -> int:
        """Return light type."""
        if self._cfg.endpoint == "light":
            return self._cfg.light_type
        return LIGHT_TYPE_NONE

    @property
    def min_mireds(self) -> int:
        """Return the coldest color_temp that this light supports."""
        return self._cfg.min_mireds

    @property
    def max_mireds(self) -> int:
        """Return the warmest color_temp that this light supports."""
        return self._cfg.max_mireds

    async def set_state(self, state: bool, attributes: dict[str, Any]) -> None:
        """Turn the light on or off."""
        if self._cfg.endpoint == "relay":
            await self._set_state_relay(state)
        else:
            await self._set_state_light(state, attributes)

    @property
    def supports_transition(self) -> bool:
        """Return if the light supports transitions."""
        return self.light_type != LIGHT_TYPE_NONE and not self._cfg.tuya

    async def _set_state_relay(self, state: bool) -> None:
        """Turn the relay on or off."""
        payload = self._cfg.state_power_on if state else self._cfg.state_power_off
        command = f"{COMMAND_POWER}{self._cfg.idx+1}"
        await self._mqtt_client.publish(
            self._cfg.command_topic + command,
            payload,
        )

    async def _set_state_light(self, state: bool, attributes: dict[str, Any]) -> None:
        idx = self._cfg.idx

        commands: list[tuple[str, str | float]] = []
        transition = attributes.get("transition", 0)
        do_transition = transition > 0

        argument: str | float

        # Set fade
        if self.supports_transition and "transition" in attributes:
            command = COMMAND_FADE
            argument = 1 if do_transition else 0
            commands.append((command, argument))

        if do_transition:
            speed = self._calculate_speed(state, attributes)

            # Clamp speed to the range 1..40
            speed = min(max(speed, 1), 40)
            command = COMMAND_SPEED
            commands.append((command, speed))

        argument = self._cfg.state_power_on if state else self._cfg.state_power_off
        command = f"{COMMAND_POWER}{idx+1}"
        if "brightness" in attributes:
            argument = attributes["brightness"]
            if self._cfg.control_by_channel:
                command = f"{COMMAND_CHANNEL}{idx+1}"
            else:
                command = self._cfg.dimmer_cmd

        commands.append((command, argument))

        if "color" in attributes:
            color = attributes["color"]
            argument = f"{color[0]},{color[1]},{color[2]}{self._cfg.color_suffix}"
            command = f"{COMMAND_COLOR}2"
            commands.append((command, argument))
        if "color_hs" in attributes:
            argument = round(attributes["color_hs"][0])
            command = "HsbColor1"  # Hue
            commands.append((command, argument))
            argument = round(attributes["color_hs"][1])
            command = "HsbColor2"  # Saturation
            commands.append((command, argument))
        if "color_temp" in attributes:
            argument = attributes["color_temp"]
            command = COMMAND_CT
            commands.append((command, argument))
        if "effect" in attributes and self.effect_list:
            try:
                effect = attributes["effect"]
                argument = self.effect_list.index(effect)
                command = COMMAND_SCHEME
                commands.append((command, argument))
            except ValueError:
                _LOGGER.debug("Unknown effect %s", effect)
        if "white_value" in attributes:
            argument = attributes["white_value"]
            command = COMMAND_WHITE
            commands.append((command, argument))

        if self._cfg.not_power_linked and "brightness" in attributes:
            # Always send power
            argument = self._cfg.state_power_on if state else self._cfg.state_power_off
            command = f"{COMMAND_POWER}{idx+1}"
            commands.append((command, argument))

        await send_commands(self._mqtt_client, self._cfg.command_topic, commands)

    def _calculate_speed(self, state: bool, attributes: dict[str, Any]) -> float:
        # Calculate speed:
        # Home Assistant's transition is the transition time in seconds.
        # Tasmota's speed command is the number of half-seconds, scaled for a 100% fade
        # if fade_fixed_duration (SetOption117) is not set
        transition: int = attributes.get("transition", 0)

        if self._cfg.fade_fixed_duration:
            # Fading at fixed duration
            return round(transition * 2)

        old_brightness = self._brightness if self._brightness is not None else 100
        now_brightness = old_brightness if self._state else 0

        new_brightness = attributes.get("brightness", old_brightness if state else 0)

        now_channels = []
        new_channels = []
        # Calculate normalized brightness for all channels
        if self.light_type >= LIGHT_TYPE_COLDWARM:
            if self.light_type >= LIGHT_TYPE_RGB and self._color:
                if "color" in attributes:
                    new_color = attributes["color"]
                elif "color_hs" in attributes:
                    # Convert hs_color to color
                    color_hs = attributes["color_hs"]
                    rgb = colorsys.hsv_to_rgb(color_hs[0] / 360, color_hs[1] / 100, 1)
                    rgb = (int(rgb[0] * 255), int(rgb[1] * 255), int(rgb[2] * 255))
                    new_color = rgb
                else:
                    new_color = self._color
                now_color = [x / 255 for x in self._color]
                new_color = [x / 255 for x in new_color]
                now_channels.extend(now_color)
                new_channels.extend(new_color)
            if self.light_type >= LIGHT_TYPE_COLDWARM and self._color_temp:
                now_color_temp = self._color_temp
                new_color_temp = attributes.get("color_temp", self._color_temp)
                mired_range = self.max_mireds - self.min_mireds
                now_ct_ratio = (now_color_temp - self.min_mireds) / mired_range
                new_ct_ratio = (new_color_temp - self.min_mireds) / mired_range
                now_channels.append(now_ct_ratio)
                new_channels.append(new_ct_ratio)
            now_channels = [x * now_brightness / 100 for x in now_channels]
            new_channels = [x * new_brightness / 100 for x in new_channels]

        # 1-channel dimmer, or color / color_temp unknown
        if not new_channels:
            new_channels = [new_brightness / 100]
            now_channels = [now_brightness / 100]

        # Scale transition to the channel with the largest brightness change
        abs_changes = map(
            abs, [x1 - x2 for (x1, x2) in zip(now_channels, new_channels)]
        )
        # Mypy is confused about the map, override the inferred typing
        if (delta_ratio := max(abs_changes)) == 0:  # type:ignore[type-var]
            speed = 0
        else:
            speed = round(transition * 2 / delta_ratio)  # type:ignore[operator]
        return speed



"""Tasmota types."""
from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from typing import Any, TypedDict

from .entity import TasmotaAvailabilityConfig, TasmotaEntityConfig


@dataclass(frozen=True, kw_only=True)
class TasmotaBaseSensorConfig(TasmotaAvailabilityConfig, TasmotaEntityConfig):
    """Tasmota Base Sensor configuration."""


DiscoveryHashType = tuple[str, str, str, str | int]
DeviceDiscoveredCallback = Callable[[dict, str], Coroutine[Any, Any, None]]
SensorsDiscoveredCallback = Callable[
    [list[tuple[TasmotaBaseSensorConfig, DiscoveryHashType]], str],
    Coroutine[Any, Any, None],
]


class TasmotaDeviceConfig(TypedDict, total=False):
    """Tasmota device config."""

    ip: str
    mac: str
    manufacturer: str
    md: str
    name: str
    sw: str



"""Tasmota sensor."""
from __future__ import annotations

from dataclasses import dataclass
import logging
from typing import Any

from .const import (
    CONCENTRATION_MICROGRAMS_PER_CUBIC_METER,
    CONCENTRATION_PARTS_PER_BILLION,
    CONCENTRATION_PARTS_PER_MILLION,
    CONF_MAC,
    CONF_SENSOR,
    DEGREE,
    ELECTRICAL_CURRENT_AMPERE,
    ELECTRICAL_VOLT_AMPERE,
    ENERGY_KILO_WATT_HOUR,
    FREQUENCY_HERTZ,
    LENGTH_CENTIMETERS,
    LIGHT_LUX,
    MASS_KILOGRAMS,
    PERCENTAGE,
    POWER_FACTOR,
    POWER_WATT,
    PRESSURE_HPA,
    PRESSURE_MMHG,
    REACTIVE_ENERGY_KILO_VOLT_AMPERE_HOUR,
    REACTIVE_POWER,
    SENSOR_ACTIVE_ENERGYEXPORT,
    SENSOR_ACTIVE_ENERGYIMPORT,
    SENSOR_ACTIVE_POWERUSAGE,
    SENSOR_AMBIENT,
    SENSOR_APPARENT_POWERUSAGE,
    SENSOR_BATTERY,
    SENSOR_CCT,
    SENSOR_CF1,
    SENSOR_CF2_5,
    SENSOR_CF10,
    SENSOR_CO2,
    SENSOR_COLOR_BLUE,
    SENSOR_COLOR_GREEN,
    SENSOR_COLOR_RED,
    SENSOR_CURRENT,
    SENSOR_CURRENTNEUTRAL,
    SENSOR_DEWPOINT,
    SENSOR_DISTANCE,
    SENSOR_ECO2,
    SENSOR_ENERGY,
    SENSOR_FREQUENCY,
    SENSOR_HUMIDITY,
    SENSOR_ILLUMINANCE,
    SENSOR_MOISTURE,
    SENSOR_PB0_3,
    SENSOR_PB0_5,
    SENSOR_PB1,
    SENSOR_PB2_5,
    SENSOR_PB5,
    SENSOR_PB10,
    SENSOR_PHASEANGLE,
    SENSOR_PM1,
    SENSOR_PM2_5,
    SENSOR_PM10,
    SENSOR_POWERFACTOR,
    SENSOR_POWERUSAGE,
    SENSOR_PRESSURE,
    SENSOR_PRESSUREATSEALEVEL,
    SENSOR_PROXIMITY,
    SENSOR_REACTIVE_ENERGYEXPORT,
    SENSOR_REACTIVE_ENERGYIMPORT,
    SENSOR_REACTIVE_POWERUSAGE,
    SENSOR_SPEED,
    SENSOR_TEMPERATURE,
    SENSOR_TODAY,
    SENSOR_TOTAL,
    SENSOR_TOTAL_START_TIME,
    SENSOR_TVOC,
    SENSOR_UNIT_PRESSURE,
    SENSOR_UNIT_SPEED,
    SENSOR_UNIT_TEMPERATURE,
    SENSOR_VOLTAGE,
    SENSOR_WEIGHT,
    SENSOR_YESTERDAY,
    SPEED_FEET_PER_SECOND,
    SPEED_KILOMETERS_PER_HOUR,
    SPEED_KNOT,
    SPEED_METERS_PER_SECOND,
    SPEED_MILES_PER_HOUR,
    SPEED_YARDS_PER_SECOND,
    TEMP_CELSIUS,
    TEMP_FAHRENHEIT,
    TEMP_KELVIN,
    VOLT,
)
from .entity import TasmotaAvailability, TasmotaEntity
from .models import DiscoveryHashType, TasmotaBaseSensorConfig
from .mqtt import ReceiveMessage
from .utils import (
    config_get_state_offline,
    config_get_state_online,
    get_topic_command_status,
    get_topic_stat_status,
    get_topic_tele_sensor,
    get_topic_tele_will,
    get_value_by_path,
)

IGNORED_SENSORS = ["Time", "PN532", "RDM6300"]

# QUANTITY                        UNIT            CLASS/ICON
# SENSOR_TEMPERATURE                              "dev_cla":"temperature"
# SENSOR_DEWPOINT                                 "ic":"mdi:weather-rainy"
# SENSOR_PRESSURE                                 "dev_cla":"pressure"
# SENSOR_PRESSUREATSEALEVEL                       "dev_cla":"pressure"
# SENSOR_APPARENT_POWERUSAGE      VA              "dev_cla":"power"
# SENSOR_BATTERY                  %               "dev_cla":"battery"
# SENSOR_CURRENT                  A               "ic":"mdi:alpha-a-circle-outline"
# SENSOR_DISTANCE                 Cm              "ic":"mdi:leak"
# SENSOR_FREQUENCY                Hz              "ic":"mdi:current-ac"
# SENSOR_HUMIDITY                 %               "dev_cla":"humidity"
# SENSOR_ILLUMINANCE              LX              "dev_cla":"illuminance"
# SENSOR_MOISTURE                 %               "ic":"mdi:cup-water"
# SENSOR_PB0_3                    ppd             "ic":"mdi:flask"
# SENSOR_PB0_5                    ppd             "ic":"mdi:flask"
# SENSOR_PB1                      ppd             "ic":"mdi:flask"
# SENSOR_PB2_5                    ppd             "ic":"mdi:flask"
# SENSOR_PB5                      ppd             "ic":"mdi:flask"
# SENSOR_PB10                     ppd             "ic":"mdi:flask"
# SENSOR_PM1                      µg/m³           "ic":"mdi:air-filter"
# SENSOR_PM2_5                    µg/m³           "ic":"mdi:air-filter"
# SENSOR_PM10                     µg/m³           "ic":"mdi:air-filter"
# SENSOR_POWERFACTOR              Cos φ           "ic":"mdi:alpha-f-circle-outline"
# SENSOR_POWERUSAGE               W               "dev_cla":"power"
# SENSOR_TOTAL_START_TIME                         "ic":"mdi:progress-clock"
# SENSOR_REACTIVE_POWERUSAGE      VAr             "dev_cla":"power"
# SENSOR_TODAY                    kWh             "dev_cla":"power"
# SENSOR_TOTAL                    kWh             "dev_cla":"power"
# SENSOR_VOLTAGE                  V               "ic":"mdi:alpha-v-circle-outline"
# SENSOR_WEIGHT                   Kg              "ic":"mdi:scale"
# SENSOR_YESTERDAY                kWh             "dev_cla":"power"
# SENSOR_CO2                      ppm             "ic":"mdi:molecule-co2"
# SENSOR_ECO2                     ppm             "ic":"mdi:molecule-co2"
# SENSOR_TVOC                     ppb             "ic":"mdi:air-filter"
# SENSOR_COLOR_RED                R               "ic":"mdi:palette"
# SENSOR_COLOR_GREEN              G               "ic":"mdi:palette"
# SENSOR_COLOR_BLUE               B               "ic":"mdi:palette"
# SENSOR_CCT                      K               "ic":"mdi:temperature-kelvin"
# SENSOR_PROXIMITY                                "ic":"mdi:ruler"
# SENSOR_AMBIENT                  LX              "dev_cla":"illuminance"


SENSOR_UNIT_MAP = {
    SENSOR_APPARENT_POWERUSAGE: ELECTRICAL_VOLT_AMPERE,
    SENSOR_BATTERY: PERCENTAGE,
    SENSOR_CURRENT: ELECTRICAL_CURRENT_AMPERE,
    SENSOR_CURRENTNEUTRAL: ELECTRICAL_CURRENT_AMPERE,
    SENSOR_DISTANCE: LENGTH_CENTIMETERS,
    SENSOR_FREQUENCY: FREQUENCY_HERTZ,
    SENSOR_HUMIDITY: PERCENTAGE,
    SENSOR_ILLUMINANCE: LIGHT_LUX,
    SENSOR_MOISTURE: PERCENTAGE,
    SENSOR_PB0_3: "ppd",
    SENSOR_PB0_5: "ppd",
    SENSOR_PB1: "ppd",
    SENSOR_PB2_5: "ppd",
    SENSOR_PB5: "ppd",
    SENSOR_PB10: "ppd",
    SENSOR_PM1: CONCENTRATION_MICROGRAMS_PER_CUBIC_METER,
    SENSOR_PM2_5: CONCENTRATION_MICROGRAMS_PER_CUBIC_METER,
    SENSOR_PM10: CONCENTRATION_MICROGRAMS_PER_CUBIC_METER,
    SENSOR_CF1: CONCENTRATION_MICROGRAMS_PER_CUBIC_METER,
    SENSOR_CF2_5: CONCENTRATION_MICROGRAMS_PER_CUBIC_METER,
    SENSOR_CF10: CONCENTRATION_MICROGRAMS_PER_CUBIC_METER,
    SENSOR_PHASEANGLE: DEGREE,
    SENSOR_POWERFACTOR: POWER_FACTOR,
    SENSOR_POWERUSAGE: POWER_WATT,
    SENSOR_TOTAL_START_TIME: None,
    SENSOR_ACTIVE_POWERUSAGE: POWER_WATT,
    SENSOR_REACTIVE_POWERUSAGE: REACTIVE_POWER,
    SENSOR_TODAY: ENERGY_KILO_WATT_HOUR,
    SENSOR_TOTAL: ENERGY_KILO_WATT_HOUR,
    SENSOR_VOLTAGE: VOLT,
    SENSOR_WEIGHT: MASS_KILOGRAMS,
    SENSOR_YESTERDAY: ENERGY_KILO_WATT_HOUR,
    SENSOR_ENERGY: ENERGY_KILO_WATT_HOUR,
    SENSOR_ACTIVE_ENERGYEXPORT: ENERGY_KILO_WATT_HOUR,
    SENSOR_ACTIVE_ENERGYIMPORT: ENERGY_KILO_WATT_HOUR,
    SENSOR_REACTIVE_ENERGYEXPORT: REACTIVE_ENERGY_KILO_VOLT_AMPERE_HOUR,
    SENSOR_REACTIVE_ENERGYIMPORT: REACTIVE_ENERGY_KILO_VOLT_AMPERE_HOUR,
    SENSOR_CO2: CONCENTRATION_PARTS_PER_MILLION,
    SENSOR_ECO2: CONCENTRATION_PARTS_PER_MILLION,
    SENSOR_TVOC: CONCENTRATION_PARTS_PER_BILLION,
    SENSOR_COLOR_RED: "R",
    SENSOR_COLOR_GREEN: "G",
    SENSOR_COLOR_BLUE: "B",
    SENSOR_CCT: TEMP_KELVIN,
    SENSOR_PROXIMITY: " ",
    SENSOR_AMBIENT: LIGHT_LUX,
}

SUPPORTED_PRESSURE_UNITS = [PRESSURE_HPA, PRESSURE_MMHG]
SUPPORTED_SPEED_UNITS = [
    SPEED_METERS_PER_SECOND,
    SPEED_KILOMETERS_PER_HOUR,
    SPEED_KNOT,
    SPEED_MILES_PER_HOUR,
    SPEED_FEET_PER_SECOND,
    SPEED_YARDS_PER_SECOND,
]
SUPPORTED_TEMPERATURE_UNITS = [TEMP_CELSIUS, TEMP_FAHRENHEIT]

SENSOR_DYNAMIC_UNIT_MAP = {
    SENSOR_DEWPOINT: (SENSOR_UNIT_TEMPERATURE, SUPPORTED_TEMPERATURE_UNITS),
    SENSOR_PRESSURE: (SENSOR_UNIT_PRESSURE, SUPPORTED_PRESSURE_UNITS),
    SENSOR_PRESSUREATSEALEVEL: (SENSOR_UNIT_PRESSURE, SUPPORTED_PRESSURE_UNITS),
    SENSOR_SPEED: (SENSOR_UNIT_SPEED, SUPPORTED_SPEED_UNITS),
    SENSOR_TEMPERATURE: (SENSOR_UNIT_TEMPERATURE, SUPPORTED_TEMPERATURE_UNITS),
}

LAST_RESET_SENSOR_MAP = {SENSOR_TOTAL: SENSOR_TOTAL_START_TIME}

_LOGGER = logging.getLogger(__name__)


@dataclass(frozen=True, kw_only=True)
class TasmotaSensorConfig(TasmotaBaseSensorConfig):
    """Tasmota Status Sensor configuration."""

    last_reset_path: list[str | int] | None
    poll_topic: str
    quantity: str
    unit: str | None
    state_topic1: str
    state_topic2: str
    value_path: list[str | int]

    @classmethod
    def from_discovery_message(
        cls,
        device_config: dict,
        sensor_config: dict,
        platform: str,
        sensor_name: str,
        value_path: list[str | int],
        parent_path: list[str | int],
        quantity: str,
    ) -> TasmotaSensorConfig:
        """Instantiate from discovery message."""
        unit = SENSOR_UNIT_MAP.get(quantity)
        if quantity in SENSOR_DYNAMIC_UNIT_MAP:
            key, supported_units = SENSOR_DYNAMIC_UNIT_MAP[quantity]
            if (unit := sensor_config[CONF_SENSOR].get(key)) not in supported_units:
                _LOGGER.warning("Unknown unit %s for %s", unit, quantity)

        if last_reset_key := LAST_RESET_SENSOR_MAP.get(quantity):
            last_reset_path = list(parent_path)
            last_reset_path.append(last_reset_key)
        else:
            last_reset_path = None

        return cls(
            endpoint="sensor",
            idx=None,
            friendly_name=sensor_name,
            last_reset_path=last_reset_path,
            mac=device_config[CONF_MAC],
            platform=platform,
            poll_payload="10",
            poll_topic=get_topic_command_status(device_config),
            availability_topic=get_topic_tele_will(device_config),
            availability_offline=config_get_state_offline(device_config),
            availability_online=config_get_state_online(device_config),
            quantity=quantity,
            state_topic1=get_topic_tele_sensor(device_config),
            state_topic2=get_topic_stat_status(device_config, 10),
            unit=unit,
            value_path=value_path,
        )

    @property
    def unique_id(self) -> str:
        """Return unique_id."""
        sensor_id = "_".join([str(i) for i in self.value_path])
        return f"{self.mac}_{self.platform}_{self.endpoint}_{sensor_id}"


class TasmotaSensor(TasmotaAvailability, TasmotaEntity):
    """Representation of Tasmota Status Sensors."""

    _cfg: TasmotaSensorConfig

    def __init__(self, **kwds: Any):
        """Initialize."""
        self._sub_state: dict | None = None
        super().__init__(**kwds)

    async def subscribe_topics(self) -> None:
        """Subscribe to topics."""

        def state_message_received(msg: ReceiveMessage) -> None:
            """Handle new MQTT state messages."""
            if not self._on_state_callback:
                return

            last_reset_path = self._cfg.last_reset_path
            if msg.topic == self._cfg.state_topic1:
                state = get_value_by_path(msg.payload, self._cfg.value_path[:-1])
                last_node = self._cfg.value_path[-1]
            if msg.topic == self._cfg.state_topic2:
                prefix: list[str | int] = ["StatusSNS"]
                value_path = prefix + self._cfg.value_path
                state = get_value_by_path(msg.payload, value_path[:-1])
                last_node = value_path[-1]
                if self._cfg.last_reset_path:
                    last_reset_path = prefix + self._cfg.last_reset_path
            if state is not None:
                # Indexed sensors may be announced with more indices than present in
                # the status. Handle this gracefully wihtout throwing. This is a
                # workaround for energy sensors which are announced with multiple phases
                # but where the actual sensor sends updates with fewer phases.
                kwargs = {}
                try:
                    if hasattr(state, "__getitem__"):
                        state = state[last_node]
                    elif last_node != 0:
                        return
                except (IndexError, KeyError):
                    return
                if last_reset_path:
                    if last_reset := get_value_by_path(msg.payload, last_reset_path):
                        kwargs["last_reset"] = last_reset
                self._on_state_callback(state, **kwargs)

        availability_topics = self.get_availability_topics()
        topics = {
            # Periodic state update (tele/Sensor)
            "state_topic1": {
                "event_loop_safe": True,
                "topic": self._cfg.state_topic1,
                "msg_callback": state_message_received,
            },
            # Polled state update (stat/STATUS10)
            "state_topic2": {
                "event_loop_safe": True,
                "topic": self._cfg.state_topic2,
                "msg_callback": state_message_received,
            },
        }
        topics = {**topics, **availability_topics}

        self._sub_state = await self._mqtt_client.subscribe(
            self._sub_state,
            topics,
        )

    async def unsubscribe_topics(self) -> None:
        """Unsubscribe to all MQTT topics."""
        self._sub_state = await self._mqtt_client.unsubscribe(self._sub_state)

    @property
    def quantity(self) -> str:
        """Return the sensor's quantity (speed, mass, etc.)."""
        return self._cfg.quantity

    @property
    def unit(self) -> str | None:
        """Return the unit this state is expressed in."""
        return self._cfg.unit


# Simple sensor:
# {"INA219":{"Voltage":4.494,"Current":0.020,"Power":0.089}}
# Array sensor:
# {"ENERGY":
#   {
#     "TotalStartTime":"2018-11-23T15:33:47",
#     "Total":0.017,
#     "TotalTariff":[0.000,0.017],
#     "Yesterday":0.000,
#     "Today":0.002,
#     "ExportActive":0.000,
#     "ExportTariff":[0.000,0.000],
#     "Period":0.00,
#     "Power":0.00,
#     "ApparentPower":7.84,
#     "ReactivePower":-7.21,
#     "Factor":0.39,
#     "Frequency":50.0,
#     "Voltage":234.31,
#     "Current":0.039,
#     "ImportActive":12.580,
#     "ImportReactive":0.002,
#     "ExportReactive":39.131,
#     "PhaseAngle":290.45}}
# Nested sensor:
# {
#   "Time":"2020-03-03T00:00:00+00:00",
#   "TX23":{
#     "Speed":{"Act":14.8,"Avg":8.5,"Min":12.2,"Max":14.8},
#     "Dir":{"Card":"WSW","Deg":247.5,"Avg":266.1,"AvgCard":"W","Min":247.5,"Max":247.5,"Range":0}
#   },
#   "SpeedUnit":"km/h"
# }
def _get_sensor_entity(
    sensor_discovery_message: dict,
    device_discovery_msg: dict,
    sensor_path: list[str | int],
    parent_path: list[str | int],
    quantity: str,
) -> tuple[TasmotaSensorConfig, DiscoveryHashType]:
    sensorname = " ".join([str(i) for i in sensor_path])
    discovery_hash = (
        device_discovery_msg[CONF_MAC],
        "sensor",
        "sensor",
        sensorname,
    )
    sensor_config = TasmotaSensorConfig.from_discovery_message(
        device_discovery_msg,
        sensor_discovery_message,
        "sensor",
        sensorname,
        sensor_path,
        parent_path,
        quantity,
    )
    return (sensor_config, discovery_hash)


def get_sensor_entities(
    sensor_discovery_message: dict, device_discovery_msg: dict
) -> list[tuple[TasmotaBaseSensorConfig, DiscoveryHashType]]:
    """Generate sensor configuration."""
    sensor_configs: list[tuple[TasmotaBaseSensorConfig, DiscoveryHashType]] = []
    for sensorkey, sensor in sensor_discovery_message[CONF_SENSOR].items():
        sensorpath = [sensorkey]
        if sensorkey in IGNORED_SENSORS or not isinstance(sensor, dict):
            continue
        for subsensorkey, subsensor in sensor.items():
            subsensorpath = list(sensorpath)
            subsensorpath.append(subsensorkey)
            quantity = subsensorkey
            if isinstance(subsensor, dict):
                # Nested sensor
                for subsubsensorkey in subsensor.keys():
                    subsubsensorpath = list(subsensorpath)
                    subsubsensorpath.append(subsubsensorkey)
                    sensor_configs.append(
                        _get_sensor_entity(
                            sensor_discovery_message,
                            device_discovery_msg,
                            subsubsensorpath,
                            subsubsensorpath[:-1],
                            quantity,
                        )
                    )
            elif isinstance(subsensor, list):
                # Array sensor
                for idx, _ in enumerate(subsensor):
                    subsubsensorpath = list(subsensorpath)
                    subsubsensorpath.append(idx)
                    sensor_configs.append(
                        _get_sensor_entity(
                            sensor_discovery_message,
                            device_discovery_msg,
                            subsubsensorpath,
                            subsensorpath[:-1],
                            quantity,
                        )
                    )
            else:
                # Simple sensor
                sensor_configs.append(
                    _get_sensor_entity(
                        sensor_discovery_message,
                        device_discovery_msg,
                        subsensorpath,
                        subsensorpath[:-1],
                        quantity,
                    )
                )

    return sensor_configs



"""Tasmota constants."""
from typing import Final

AUTOMATION_TYPE_TRIGGER: Final = "trigger"

COMMAND_BACKLOG: Final = "Backlog"
COMMAND_CHANNEL: Final = "Channel"
COMMAND_COLOR: Final = "Color"
COMMAND_CT: Final = "CT"
COMMAND_DIMMER: Final = "Dimmer"
COMMAND_FADE: Final = "Fade2"
COMMAND_FANSPEED: Final = "FanSpeed"
COMMAND_POWER: Final = "Power"
COMMAND_SCHEME: Final = "Scheme"
COMMAND_SHUTTER_CLOSE: Final = "ShutterClose"
COMMAND_SHUTTER_OPEN: Final = "ShutterOpen"
COMMAND_SHUTTER_POSITION: Final = "ShutterPosition"
COMMAND_SHUTTER_STOP: Final = "ShutterStop"
COMMAND_SHUTTER_TILT: Final = "ShutterTilt"
COMMAND_SPEED: Final = "Speed2"
COMMAND_WHITE: Final = "White"

CONF_BUTTON: Final = "btn"
CONF_DEVICENAME: Final = "dn"
CONF_FRIENDLYNAME: Final = "fn"
CONF_FULLTOPIC: Final = "ft"
CONF_IFAN: Final = "if"
CONF_IP: Final = "ip"
CONF_HOSTNAME: Final = "hn"
CONF_MAC: Final = "mac"
CONF_LIGHT_SUBTYPE: Final = "lt_st"
CONF_LINK_RGB_CT: Final = "lk"  # RGB + white channels linked to a single light
CONF_MODEL: Final = "md"
CONF_OFFLINE: Final = "ofln"
CONF_ONLINE: Final = "onln"
CONF_OPTIONS: Final = "so"
CONF_PREFIX: Final = "tp"
CONF_SENSOR: Final = "sn"
CONF_BATTERY: Final = "bat"
CONF_SHUTTER_OPTIONS: Final = "sho"
CONF_SHUTTER_TILT: Final = "sht"
CONF_STATE: Final = "state"
CONF_RELAY: Final = "rl"
CONF_SW_VERSION: Final = "sw"
CONF_SWITCH: Final = "swc"
CONF_SWITCHNAME: Final = "swn"
CONF_TOPIC: Final = "t"
CONF_TUYA: Final = "ty"
CONF_VERSION: Final = "ver"

CONF_MANUFACTURER: Final = "manufacturer"
CONF_NAME: Final = "name"

FAN_SPEED_OFF: Final = 0
FAN_SPEED_LOW: Final = 1
FAN_SPEED_MEDIUM: Final = 2
FAN_SPEED_HIGH: Final = 3

LST_NONE: Final = 0
LST_SINGLE: Final = 1
LST_COLDWARM: Final = 2
LST_RGB: Final = 3
LST_RGBW: Final = 4
LST_RGBCW: Final = 5

# fmt: off
OPTION_MQTT_RESPONSE: Final = "4"          # Return MQTT response as RESULT or %COMMAND%
OPTION_BUTTON_SWAP: Final = "11"           # Swap button single and double press functionality
OPTION_BUTTON_SINGLE: Final = "13"         # Allow immediate action on single button press
OPTION_DECIMAL_TEXT: Final = "17"          # Show Color string as hex or comma-separated
OPTION_NOT_POWER_LINKED: Final = "20"      # Update of Dimmer/Color/CT without turning power on
OPTION_HASS_LIGHT: Final = "30"            # Enforce Home Assistant auto-discovery as light
OPTION_PWM_MULTI_CHANNELS: Final = "68"    # Multi-channel PWM instead of a single light
OPTION_MQTT_BUTTONS: Final = "73"          # Enable Buttons decoupling and send multi-press and hold MQTT messages
OPTION_SHUTTER_MODE: Final = "80"          # Blinds and shutters support; removed in Tasmota 9.0.0.4
OPTION_REDUCED_CT_RANGE: Final = "82"      # Reduce the CT range from 153..500 to 200.380
OPTION_MQTT_SWITCHES: Final = "114"        # Enable sending switch MQTT messages
OPTION_FADE_FIXED_DURATION: Final = "117"  # Run fading at fixed duration instead of fixed slew rate
# fmt: on

PREFIX_CMND: Final = 0
PREFIX_STAT: Final = 1
PREFIX_TELE: Final = 2

RL_NONE: Final = 0
RL_RELAY: Final = 1
RL_LIGHT: Final = 2
RL_SHUTTER: Final = 3

RSLT_ACTION: Final = "Action"
RSLT_POWER: Final = "POWER"
RSLT_SHUTTER: Final = "Shutter"
RSLT_STATE: Final = "STATE"
RSLT_TRIG: Final = "TRIG"

SENSOR_ATTRIBUTE_RSSI: Final = "RSSI"
SENSOR_ATTRIBUTE_UPTIME: Final = "Uptime"
SENSOR_ATTRIBUTE_SIGNAL: Final = "Signal"
SENSOR_ATTRIBUTE_WIFI_LINKCOUNT: Final = "LinkCount"
SENSOR_ATTRIBUTE_WIFI_DOWNTIME: Final = "Downtime"
SENSOR_ATTRIBUTE_MQTTCOUNT: Final = "MqttCount"

SENSOR_TEMPERATURE: Final = "Temperature"
SENSOR_DEWPOINT: Final = "DewPoint"
SENSOR_PRESSURE: Final = "Pressure"
SENSOR_PRESSUREATSEALEVEL: Final = "SeaPressure"
SENSOR_APPARENT_POWERUSAGE: Final = "ApparentPower"
SENSOR_BATTERY: Final = "Battery"
SENSOR_CURRENT: Final = "Current"
SENSOR_CURRENTNEUTRAL: Final = "CurrentNeutral"
SENSOR_DISTANCE: Final = "Distance"
SENSOR_FREQUENCY: Final = "Frequency"
SENSOR_HUMIDITY: Final = "Humidity"
SENSOR_ILLUMINANCE: Final = "Illuminance"
SENSOR_MOISTURE: Final = "Moisture"
SENSOR_PB0_3: Final = "PB0.3"
SENSOR_PB0_5: Final = "PB0.5"
SENSOR_PB1: Final = "PB1"
SENSOR_PB2_5: Final = "PB2.5"
SENSOR_PB5: Final = "PB5"
SENSOR_PB10: Final = "PB10"
SENSOR_PM1: Final = "PM1"
SENSOR_PM2_5: Final = "PM2.5"
SENSOR_PM10: Final = "PM10"
SENSOR_CF1: Final = "CF1"
SENSOR_CF2_5: Final = "CF2.5"
SENSOR_CF10: Final = "CF10"
SENSOR_PHASEANGLE: Final = "PhaseAngle"
SENSOR_POWERFACTOR: Final = "Factor"
SENSOR_POWERUSAGE: Final = "Power"
SENSOR_SPEED: Final = "Speed"
SENSOR_TOTAL_START_TIME: Final = "TotalStartTime"
SENSOR_ACTIVE_POWERUSAGE: Final = "ActivePower"
SENSOR_REACTIVE_POWERUSAGE: Final = "ReactivePower"
SENSOR_TODAY: Final = "Today"
SENSOR_TOTAL: Final = "Total"
SENSOR_VOLTAGE: Final = "Voltage"
SENSOR_WEIGHT: Final = "Weight"
SENSOR_YESTERDAY: Final = "Yesterday"
SENSOR_ENERGY: Final = "Energy"
SENSOR_ACTIVE_ENERGYEXPORT: Final = "ExportActive"
SENSOR_ACTIVE_ENERGYIMPORT: Final = "ImportActive"
SENSOR_REACTIVE_ENERGYEXPORT: Final = "ExportReactive"
SENSOR_REACTIVE_ENERGYIMPORT: Final = "ImportReactive"
SENSOR_CO2: Final = "CarbonDioxide"
SENSOR_ECO2: Final = "eCO2"
SENSOR_TVOC: Final = "TVOC"
SENSOR_COLOR_RED: Final = "Red"
SENSOR_COLOR_GREEN: Final = "Green"
SENSOR_COLOR_BLUE: Final = "Blue"
SENSOR_CCT: Final = "CCT"
SENSOR_PROXIMITY: Final = "Proximity"
SENSOR_AMBIENT: Final = "Ambient"
SENSOR_SWITCH: Final = "Switch"
SENSOR_STATUS_IP: Final = "status_ip"
SENSOR_STATUS_LAST_RESTART_TIME: Final = "last_restart_time"
SENSOR_STATUS_LINK_COUNT: Final = "status_link_count"
SENSOR_STATUS_MQTT_COUNT: Final = "status_mqtt_count"
SENSOR_STATUS_RESTART_REASON: Final = "status_restart_reason"
SENSOR_STATUS_RSSI: Final = "status_rssi"
SENSOR_STATUS_SIGNAL: Final = "status_signal"
SENSOR_STATUS_SSID: Final = "status_ssid"
SENSOR_STATUS_VERSION: Final = "status_version"

SENSOR_STATUS_BATTERY_PERCENTAGE: Final = "status_battery_percentage"

SENSOR_UNIT_PRESSURE: Final = "PressureUnit"
SENSOR_UNIT_SPEED: Final = "SpeedUnit"
SENSOR_UNIT_TEMPERATURE: Final = "TempUnit"

SHUTTER_DIRECTION: Final = "Direction"
SHUTTER_DIRECTION_DOWN: Final = -1
SHUTTER_DIRECTION_STOP: Final = 0
SHUTTER_DIRECTION_UP: Final = 1
SHUTTER_POSITION: Final = "Position"
SHUTTER_TILT: Final = "Tilt"

SHUTTER_OPTION_INVERT: Final = 1

# #### UNITS OF MEASUREMENT ####
# Power units
POWER_WATT: Final = "W"
POWER_FACTOR: Final = "Cos φ"
REACTIVE_POWER = "VAr"

# Voltage units
VOLT: Final = "V"

# Energy units
ENERGY_WATT_HOUR: Final = f"{POWER_WATT}h"
ENERGY_KILO_WATT_HOUR: Final = f"k{ENERGY_WATT_HOUR}"
REACTIVE_ENERGY_VOLT_AMPERE_HOUR = f"{REACTIVE_POWER}h"
REACTIVE_ENERGY_KILO_VOLT_AMPERE_HOUR = f"k{REACTIVE_ENERGY_VOLT_AMPERE_HOUR}"

# Electrical units
ELECTRICAL_CURRENT_AMPERE: Final = "A"
ELECTRICAL_VOLT_AMPERE: Final = f"{VOLT}{ELECTRICAL_CURRENT_AMPERE}"

# Temperature units
TEMP_CELSIUS: Final = "C"
TEMP_FAHRENHEIT: Final = "F"
TEMP_KELVIN: Final = "K"

# Degree units
DEGREE: Final = "°"

# Time units
TIME_SECONDS: Final = "s"
TIME_HOURS: Final = "h"

# Length units
LENGTH_CENTIMETERS: Final = "cm"
LENGTH_METERS: Final = "m"
LENGTH_KILOMETERS: Final = "km"

# Frequency units
FREQUENCY_HERTZ: Final = "Hz"

# Pressure units
PRESSURE_HPA: Final = "hPa"
PRESSURE_MMHG: Final = "mmHg"

# Volume units
VOLUME_CUBIC_METERS: Final = f"{LENGTH_METERS}³"

# Mass units
MASS_KILOGRAMS: Final = "kg"
MASS_MICROGRAMS: Final = "µg"

# Light units
LIGHT_LUX: Final = "lux"

# Percentage units
PERCENTAGE: Final = "%"

# Concentration units
CONCENTRATION_MICROGRAMS_PER_CUBIC_METER: Final = (
    f"{MASS_MICROGRAMS}/{VOLUME_CUBIC_METERS}"
)
CONCENTRATION_PARTS_PER_MILLION: Final = "ppm"
CONCENTRATION_PARTS_PER_BILLION: Final = "ppb"

# Speed units
SPEED_METERS_PER_SECOND: Final = f"{LENGTH_METERS}/{TIME_SECONDS}"
SPEED_KILOMETERS_PER_HOUR: Final = f"{LENGTH_KILOMETERS}/{TIME_HOURS}"
SPEED_KNOT: Final = "kn"
SPEED_MILES_PER_HOUR: Final = "mph"
SPEED_FEET_PER_SECOND: Final = "ft/s"
SPEED_YARDS_PER_SECOND: Final = "yd/s"

# Signal_strength units
SIGNAL_STRENGTH_DECIBELS: Final = "dB"
SIGNAL_STRENGTH_DECIBELS_MILLIWATT: Final = "dBm"

STATE_OFF: Final = 0
STATE_ON: Final = 1
STATE_TOGGLE: Final = 2
STATE_HOLD: Final = 3

STATUS_SENSOR: Final = "StatusSNS"

SWITCHMODE_NONE: Final = -1
SWITCHMODE_TOGGLE: Final = 0
SWITCHMODE_FOLLOW: Final = 1
SWITCHMODE_FOLLOW_INV: Final = 2
SWITCHMODE_PUSHBUTTON: Final = 3
SWITCHMODE_PUSHBUTTON_INV: Final = 4
SWITCHMODE_PUSHBUTTONHOLD: Final = 5
SWITCHMODE_PUSHBUTTONHOLD_INV: Final = 6
SWITCHMODE_PUSHBUTTON_TOGGLE: Final = 7
SWITCHMODE_TOGGLEMULTI: Final = 8
SWITCHMODE_FOLLOWMULTI: Final = 9
SWITCHMODE_FOLLOWMULTI_INV: Final = 10
SWITCHMODE_PUSHHOLDMULTI: Final = 11
SWITCHMODE_PUSHHOLDMULTI_INV: Final = 12
SWITCHMODE_PUSHON: Final = 13
SWITCHMODE_PUSHON_INV: Final = 14
SWITCHMODE_PUSH_IGNORE: Final = 15
SWITCHMODE_PUSH_IGNORE_INV: Final = 16



"""Tasmota MQTT."""
from __future__ import annotations

import asyncio
from collections.abc import Callable, Coroutine
from dataclasses import dataclass
import logging
from typing import Any

from .const import COMMAND_BACKLOG

DEBOUNCE_TIMEOUT = 1

_LOGGER = logging.getLogger(__name__)


class Timer:
    """Simple timer."""

    def __init__(
        self, timeout: float, callback: Callable[[], Coroutine[Any, Any, None]]
    ):
        self._timeout = timeout
        self._callback = callback
        self._task = asyncio.ensure_future(self._job())

    async def _job(self) -> None:
        await asyncio.sleep(self._timeout)
        await self._callback()

    def cancel(self) -> None:
        """Cancel the timer."""
        self._task.cancel()


PublishPayloadType = str | bytes | int | float | None
ReceivePayloadType = str | bytes


@dataclass(frozen=True)
class PublishMessage:
    """MQTT Message."""

    topic: str
    payload: PublishPayloadType
    qos: int | None
    retain: bool | None


@dataclass(frozen=True)
class ReceiveMessage:
    """MQTT Message."""

    topic: str
    payload: ReceivePayloadType
    qos: int
    retain: bool


class TasmotaMQTTClient:
    """Helper class to sue an external MQTT client."""

    def __init__(
        self,
        publish: Callable[
            [str, PublishPayloadType, int | None, bool | None],
            Coroutine[Any, Any, None],
        ],
        subscribe: Callable[[dict | None, dict], Coroutine[Any, Any, dict]],
        unsubscribe: Callable[[dict | None], Coroutine[Any, Any, dict]],
    ):
        """Initialize."""
        self._pending_messages: dict[PublishMessage, Timer] = {}
        self._publish = publish
        self._subscribe = subscribe
        self._unsubscribe = unsubscribe

    async def publish(
        self,
        topic: str,
        payload: PublishPayloadType,
        qos: int | None = 0,
        retain: bool | None = False,
    ) -> None:
        """Publish a message."""
        return await self._publish(topic, payload, qos, retain)

    async def publish_debounced(
        self,
        topic: str,
        payload: PublishPayloadType,
        qos: int | None = 0,
        retain: bool | None = False,
    ) -> None:
        """Publish a message, with debounce."""
        msg = PublishMessage(topic, payload, qos, retain)

        async def publish_callback() -> None:
            _LOGGER.debug("publish_debounced: publishing %s", msg)
            self._pending_messages.pop(msg)
            await self.publish(msg.topic, msg.payload, qos=msg.qos, retain=msg.retain)

        if msg in self._pending_messages:
            timer = self._pending_messages.pop(msg)
            timer.cancel()
        timer = Timer(DEBOUNCE_TIMEOUT, publish_callback)
        self._pending_messages[msg] = timer

    async def subscribe(self, sub_state: dict | None, topics: dict) -> dict:
        """Subscribe to topics."""
        return await self._subscribe(sub_state, topics)

    async def unsubscribe(self, sub_state: dict | None) -> dict:
        """Unsubscribe from topics."""
        return await self._unsubscribe(sub_state)


async def send_commands(
    mqtt_client: TasmotaMQTTClient,
    command_topic: str,
    commands: list[tuple[str, str | float]],
) -> None:
    """Send a sequence of commands."""
    backlog_topic = command_topic + COMMAND_BACKLOG
    backlog = ";".join([f"NoDelay;{command[0]} {command[1]}" for command in commands])
    await mqtt_client.publish(backlog_topic, backlog)



from setuptools import setup, find_packages


long_description = open("README.md").read()

setup(
    name="HATasmota",
    version="0.7.1",
    license="MIT",
    url="https://github.com/emontnemery/hatasmota",
    author="",
    author_email="",
    description="Python module to help parse and construct Tasmota MQTT messages.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    install_requires=list(val.strip() for val in open("requirements.txt")),
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires='>=3.6',
)
